#include <stdlib.h>
#include <timer.h>
#include <display.h>
#include <keyboard.h>

//---
//	Image management
//---

typedef struct {
	int x, y;
	int width, height;
} image_t;

#define snowflake(frame)	(image_t){ 17 * frame, 0, 17, 17 }
#define star(frame)		(image_t){ 23 * frame, 20, 23, 23 }
#define merry()			(image_t){ 3, 47, 44, 14 }
#define	christmas()		(image_t){ 49, 46, 77, 14 }

void centerimg(int x, int y, image_t img)
{
	extern Image graphics;
	dimage_part(x - img.width / 2, y - img.height / 2, &graphics, img.x,
		img.y, img.width, img.height);
}

void drawimg(int x, int y, image_t img)
{
	extern Image graphics;
	dimage_part(x, y, &graphics, img.x, img.y, img.width, img.height);
}

//---
//	Animation information
//---

int time = 0;
#define	TIME_REVERSE1	50
#define	TIME_REVERSE2	150
#define	TIME_MAX	170

typedef struct {
	int x, y;
	int final_frame;
	int frame;
} snowflake_t;

snowflake_t snow[8];
int snow_count = 0;

int wind_power = 0;
int wind_speed = 0;

//---
//	Displaying every frame
//---

void draw(void)
{
	dclear();

	if(time >= TIME_MAX && !snow_count)
	{
		dtext(30, 20, "EXE   Again");
		dtext(30, 28, "MENU  Quit");
		dtext(24, 36, "Made by lephe");
	}

	// Draw snow.
	for(int i = 0; i < snow_count; i++)
		centerimg(snow[i].x / 10, snow[i].y, snowflake(snow[i].frame));

	// Display "Merry Christmas".
	if(time >= 65 && time <= TIME_REVERSE2)
	{
		if(time <= 69)
		{
			drawimg(time - 57, 23, merry());
			drawimg(time - 61, 35, christmas());
		}
		else
		{
			drawimg(12, 23, merry());
			drawimg(8, 35, christmas());
		}
	}

	// Draw the star.
	if(time >= 80 && time <= TIME_REVERSE2)
	{
		int f = ((time - 80) / 2) % 4;
		if(f == 3) f = 1;
		centerimg(100, 21, star(f));
	}

	// Reverse some parts of the screen, if required.
	for(int i = 0; i < 16; i++)
	{
		int nb_frames = TIME_REVERSE2 - TIME_REVERSE1;
		int y = -8 * ((TIME_REVERSE2 - time) + i);
		dreverse_area(16 * i, y, 16 * i + 15, y + 8 * (nb_frames));
	}

	dupdate();
}

//---
//	Main program
//---

void initialize(void)
{
	time = 0;
	snow_count = 0;
}

void timer_handler(void)
{
	draw();
	if(time >= TIME_MAX && !snow_count) return;
	time++;

	// Adjusting wind power.
	if(wind_speed < -10) wind_power = 1;
	else if(wind_speed > 10) wind_power = -1;
	else wind_power += (rand() % 3) - 1;
	wind_speed += wind_power;

	// Moving snowflakes (effect of wind and gravity).
	for(int i = 0; i < snow_count; i++)
	{
		snow[i].x += wind_speed + (rand() % 3) - 1;
		snow[i].y += 1 + (rand() % 2);
		snow[i].frame += (snow[i].frame < snow[i].final_frame);
	}

	// Removing fallen snowflakes.
	for(int i = 0; i < snow_count; i++) if(snow[i].y > 70)
	{
		for(int j = i + 1; j < snow_count; j++)
			snow[j - 1] = snow[j];
		snow_count--;
	}

	// Creating some snow if required.
	if(time < TIME_MAX && snow_count < 8 && !(rand() % 7))
	{
		snow[snow_count].x = 160 + (rand() % 960);
		snow[snow_count].y = rand() % 16;
		snow[snow_count].frame = 0;
		snow[snow_count].final_frame = 3 + rand() % 3;
		snow_count++;
	}
}

int main(void)
{
	int key;

	timer_start(TIMER_USER, clock_setting(10, Clock_Hz), TIMER_Po_4,
		timer_handler, 0);

	do
	{
		initialize();

		do key = getkey();
		while(time < TIME_MAX || snow_count > 0 ||
			(key != KEY_MENU && key != KEY_EXE
			&& key != KEY_EXIT));
	}
	while(key == KEY_EXE);

	return EXIT_SUCCESS;
}
